
// Gives us access to all functions in the 'system' namespace
using System;
using kill = System.GC; // namespace alias

class ASimpleClass {

	// This method, being static, pertains to the
	// class and not any instances of the class.
	// C# will set initial class values sensibly if you
	// don't, so in this case, zero is appropriate
	// as the value is an int
	private static int numberOfInstances;
	
	// method used to demonstrate the keyword 'this'
	// see the function 'ChangeX'
	private int x;
	
	// A constant member, which can never be set
	public const string Constant = "can't be changed";
	public readonly string ChangeMeOnce = "can be changed once";

	// Constructor
	public ASimpleClass(){
	
		// Every time the constructor is called
		// the number of instances value is
		// increased
		numberOfInstances++;
		
		// Demonstrate constant and readonly members
		// Uncomment to see errors
		// read only members should be set in
		// the constructor
		
		//Constant = "this is illegal";
		ChangeMeOnce = "this is legal";
		//ChangeMeOnce = "but now it's not";
		
	}
	
	// Destructor
	~ASimpleClass(){
		Console.Write("\nGoodbye from {0} ", this.ToString() + " " + ASimpleProperty);
	}
	
	// Properties replace the need to write the following types
	// of methods that
	// --------------------------------------------------------->>>
	public void ChangeX(int x){
		// At this point we have a parameter being
		// passed in called x, but the object also a propery
		// called x.  How do we differentiate?
		// The 'this' keyword refers to the current instance
		// of this class for which this method is being called
		// so we simply write:
		this.x = x;
	}
	
	public int GetX(){
		// the following is equivalent 
		// to writing 'return this.x'
		return x;
	}
	// <<<---------------------------------------------------------

	// HiddenValue is the actual
	// private value that the 'property' maps to
	private int HiddenValue;
	
	// This is the actual property: it implements
	// a public way of changing and getting the value
	// that it maps to.  Anyone using this class can't
	// see 'HiddenValue'.  They can only see the publicly
	// defined property 'ASimplePropety'
	public int ASimpleProperty
	{
		get
		{
			return HiddenValue;
		}
		set
		{
			// note 'value' is C#'s representation
			// of the value passed into the 'set' method
			HiddenValue = value;
		}
	}
	
	// Now we will define another property based on
	// the static field which tells us how many instances there are.
	// Obviously we don't want this value to be written to,
	// so we make it read-only
	// Notice how we have to define the property as static
	// due to the data it maps onto
	
	public static int NumberOfInstances
	{
		get
		{
			return numberOfInstances;
		}
	}
}

class GettingStuckIntoClasses {
	
	// constructor
	public GettingStuckIntoClasses(){
		
	}	
	
	// Destructor
	~GettingStuckIntoClasses(){
		Console.Write("\nGoodbye from {0}", this.ToString());
	}
	
	// Take an instance of ASimpleClass and change its property
	public static void ChangeValue(ref ASimpleClass obj, int newValue) {
	
		obj.ASimpleProperty = newValue;
	
	}
	
	public static int Main(){
	
			// a method contained in System.GC
			// tells all destructors to execute on main's exit
			kill.RequestFinalizeOnShutdown();
	
			// begin by wriitng newline to console
			Console.WriteLine();

			// Before we create any instances of ASimpleClass, let's test
			// the 'number of instances' property to ensure it is correct
			Console.Write("Number of ASimpleClass instances is: {0}\n", ASimpleClass.NumberOfInstances);

			// create an instance of the 'ASimpleClass' class
			ASimpleClass objectInstance = new ASimpleClass();
			
			// change its simple property value by feeding the instance and the
			// new value into the static method 'ChangeValue' of our main class.
			GettingStuckIntoClasses.ChangeValue(ref objectInstance, 4);
			
			// Print that value to the screen
			Console.Write("First instance property equals {0}\n", objectInstance.ASimpleProperty);
			
			// Now use the 'number of instances' property to determine how many instances
			// of ASimpleClass are active
			Console.Write("Number of ASimpleClass instances is: {0}\n", ASimpleClass.NumberOfInstances);
			
			// create another instance of the 'ASimpleClass' class
			ASimpleClass objectInstance2 = new ASimpleClass();
			
			// change its simple property value by feeding the instance and the
			// new value into the static method 'ChangeValue' of our main class.
			GettingStuckIntoClasses.ChangeValue(ref objectInstance2, 8);
			
			// Print that value to the screen
			Console.Write("Second instance property equals {0}\n", objectInstance2.ASimpleProperty);
			
			// Now use the 'number of instances' property to determine how many instances
			// of ASimpleClass are active
			Console.Write("Number of ASimpleClass instances is: {0}\n", ASimpleClass.NumberOfInstances);
			
			// finally demonstrate the 'this' keyword by changing and printing its
			// 'x' value
			objectInstance2.ChangeX(7);
			Console.Write("'X' is now: {0}", objectInstance2.GetX());
		
			return 1; // any int value will do
 
	}
}
